/*
** $Id: process.c,v 1.3 2020/07/11 22:12:13 lpitcher Exp $
*/

#include <sys/file.h>
#include <syslog.h>

#include "logging.h"
#include "process.h"
#include "sem_util.h"
#include "exec_shell.h"

/*
** doPrep()	fork off child process to execute "preprocessing" function  of specified resource
**
** Accepts:	a char pointer to resource name string
**		an int resource lock fd
**
** Updates:	syslog with child startup, stdout & stderr, termination messages
**		related lock file (locked & unlocked) while synchronizing preprocessing function
**		related semaphore file with updated "in use" count 
**
** Returns:	PROC_WORKED if preprocessing function completed successfully, or
**		PROC_FAILED if preprocessing function failed, or doPrep() failed to invoke function
*/
int doPrep(char *resource, int semaphorefd)
{
  int rc = PROC_WORKED;
  int count;

  flock(semaphorefd,LOCK_EX);

    switch (count = getCount(semaphorefd))
    {
      case COUNTERROR:
        syslog(SYSLOG_LEVEL_MESSAGE,"PREP %s: BLOCKED by previous error",
               resource);
        rc = PROC_FAILED;
        break;

      case 0:
        if (doExit(resource,"PREP") != 0)
        {
          putBuffer(semaphorefd,"ABORT");
          rc = PROC_FAILED;
        }
        else putCount(semaphorefd,count+1);
        break;

      default:
        putCount(semaphorefd,count+1);
        break;
    }
    flock(semaphorefd,LOCK_UN);

  return rc;
}

/*
** doPost()	fork off child process to execute "postprocessing" function of specified resource
**
** Accepts:	a char pointer to resource name
**		a int value for semaphore file fd
**
** Updates:	syslog with child startup, stdout & stderr, termination messages
**		related lock file (locked & unlocked) while synchronizing postprocessing function
**		related semaphore file with updated "in use" count 
**
** Returns:	PROC_WORKED if postprocessing function completed successfully, or
**		PROC_FAILED if postprocessing function failed, or doPost() failed to invoke function
*/
int doPost(char *resource, int semaphorefd)
{
  int rc = PROC_WORKED;
  int count;

  flock(semaphorefd,LOCK_EX);
  switch (count = getCount(semaphorefd))
  {
    case COUNTERROR:
      syslog(SYSLOG_LEVEL_MESSAGE,"POST %s: BLOCKED by previous error",
             resource);
      rc = PROC_FAILED;
      break;

    case 1:
      if (doExit(resource,"POST") != 0) rc = PROC_FAILED;
      /*FALLTHRU*/
    default:
      putCount(semaphorefd,count-1);
      break;
    }
    flock(semaphorefd,LOCK_UN);

  return rc;
}

/*
** Copyright Lew Pitcher (lew.pitcher@digitalfreehold.ca) February, 2013
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License along
** with this program (see the licence.txt file); if not, write to the Free
** Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
** 02110-1301 USA.
*/


